;*************************************************************************
;		A Cooley-Tukey Radix-2 DIF FFT
;
;	Radix-2 implementation
;	Decimation In Frequency
;	Single Butterfly
;	Table Lookup of Twiddle Factors 
;	Complex Input & Complex Output
;
;	All data is assumed to be 16 bits and the intermediate
;	results are stored in 32 bits
;
;	Length Of FFT must be a Power Of 2
;	Max Length Possible is 2**15
;
;	The input/output complex data is organized as a single array
;	of real data followed by imaginary data
;		Data is stored in External Memory and is accessed by
;	TABLRD & TABLWT Instructions
;
;*************************************************************************
;
	include "17c42.h"
;
	include "17c42.mac"

;*****************************************************************************
; 	RLC16AB
;
; DESCRIPTION:
; 	16 bit rotate left A into B
;
; ARGUMENTS:
; 	2*a => b
;
; TIMING (in cycles):
;	3
;
;

RLC16AB	MACRO	a,b

	BCF	_carry
	RLCF	a+B0,W
	MOVWF	b+B0
	RLCF	a+B1,W
	MOVWF	b+B1

	ENDM

;*****************************************************************************
; 	TBLADDR
;
; DESCRIPTION:
; 	Load 16 bit table pointer with specified label
;
; TIMING (in cycles):
;	4
;

TBLADDR	MACRO	label

	MOVLW	(label)
	MOVWF	tblptrl
	MOVLW	page	(label)
	MOVWF	tblptrh

	ENDM

;*****************************************************************************
; 	ADDLBL
;
; DESCRIPTION:
; 	Add A Label (16 bit constant) To A File Register (16 bit)
;
; TIMING (in cycles):
;	4
;

ADDLBL	MACRO	label,f

	MOVLW	(label)
	ADDWF	f+B0
	MOVLW	page	(label)
	ADDWFC	f+B1

	ENDM

;*****************************************************************************

;
FftLen	.set	256	; FFT Length
Power	.set	8	; (2**Power = FftLen)
DigitRevCount	.set	239	; (FftLen-1) - (2**((Power+1)/2))

SCALE_BUTTERFLY	.set	TRUE		; intermediate scaling performed

EXT_RAM_START_ADDR	.set	0x0800	; External Memory Data Storage Start Addr
;
;*************************************************************************
;
	CBLOCK	0
		B0,B1,B2,B3	; RAM offset constants
	ENDC
;
	CBLOCK  0x18
		AARG,AARG1		; 16 bit multiplier A
		BARG,BARG1		; 16 bit multiplicand B
		DPX,DPX1,DPX2,DPX3      ; 32 bit multiplier result = A*B
	ENDC
;
	CBLOCK
		ACC, ACC1, ACC2, ACC3	; 32 bit accumulator for computations
	ENDC
;
	CBLOCK
		count1,count11		; N1
		count2,count22		; N2
		QuartLen,QuartLen1	; FftLen/4
	ENDC
;
	CBLOCK
		TF_Offset,TF_Offset1
		TF_Addr,TF_Addr1	; twiddle factor address computations
		Cos,Cos1,
		Sin,Sin1
	ENDC
;
	CBLOCK
		VarIloop,VarIloop1
		VarJloop,VarJloop1
		VarKloop
		VarL,VarL1
	ENDC
;
	CBLOCK
		Xi,Xi1
		Yi,Yi1
		Xl,Xl1
		Yl,Yl1
	ENDC
;
	CBLOCK
		Xt,Xt1
		Yt,Yt1
	ENDC

	CBLOCK
		temp,temp1,temp2,temp3
		testCount,testCount1
		PulseCount, PulseCount1
	ENDC
;
;
;******************************************************************
;		Test Program For FFT Subroutine
;******************************************************************
;
	ORG	0x0000
;
	include	"square.asm"		; Generate Test Vector Data
;
	call	R2FFT		; Compute Fourier Transform
	call	Unscramble	; Digit Reverse the scrambled data
;
; Fourier Transform Completed
;
; capture data to PIC-MASTER Trace Buffer
	MOVK16	FftLen*2,temp
	TBLADDR	ExtRamAddr	; load table pointers with data start addr
capture	
	tablrd	0,1,Sin		; table latch = mem(tblptr)
	DEC16	temp
	TFSZ16	temp
	goto	capture
;
self	goto	self
;
write
	tlwt	0,Xi
	tablwt	1,1,Xi+B1		; auto increment for Imag Data
	tlwt	0,Yi
	tablwt	1,1,Yi+B1
	return
;
;******************************************************************
;			RADIX-2 FFT
;
;	Decimation In Frequency 
;
; Input Data should be unscrambled
; Output Data at the end is in scrambled form
; To obtain the unscrambled form, the digit reverse counter 
; subroutine, "Unscramble" should be called (see the example)
;
;******************************************************************

R2FFT
	MOVK16	FftLen,count2		; count2 = N
	MOVK16	FftLen/4,QuartLen	; QuartLen = FftLen/4
	clrf	TF_Offset+B1
	MOVK	1,TF_Offset		; Init TF_Offset = 1
	MOVK	Power,VarKloop		; Kloop
Kloop					; for K = 1 to Power-1
	MOV16	count2,count1		; count1 = count2
	RRC16	count2			; count2 = count2/2
	CLR16	VarJloop		; J = 0
	CLR16	TF_Addr			; TF_Addr = 0
Jloop
;
; Read Twiddle factors from Sine/Cosine Table from Prog Mem
;
	MOVFP16	TF_Addr,tblptrl		; load sine table address to table pointers
	ADDLBL	SineTable,tblptrl	; add table offset

	tablrd	0,0,Sin			; Read Sine Value from lookup table
	tlrd	0,Sin
	tlrd	1,Sin+B1
	ADD16	QuartLen,tblptrl
	tablrd	0,0,Cos			; Read Cosine Value from table
	tlrd	0,Cos
	tlrd	1,Cos+B1
;
	ADD16	TF_Offset,TF_Addr	; TF_Addr = TF_Addr + TF_Offset
;
	RLC16AB	VarJloop,VarIloop	; I = J*2 since Real followed by Imag Data
;
Iloop

	RLC16AB	count2,VarL		; VarL = count2*2
	ADD16	VarIloop,VarL		; VarL = (I+count2)*2
;
; Get Real & Imag Data from external RAMs (Program Memory)
; load table pointers with data start addr
;
	MOVFP16	VarL,tblptrl		; read data(L)
	ADDLBL	ExtRamAddr,tblptrl	; add data addr offset

	tablrd	0,1,Xl		; auto increment for Imag Data
	tlrd	0,Xl
	tlrd	1,Xl+B1		; real data XL
	tablrd	0,0,Yl
	tlrd	0,Yl
	tlrd	1,Yl+B1		; imag data YL 

	MOVFP16	VarIloop,tblptrl	; read data(I)
	ADDLBL	ExtRamAddr,tblptrl	; add data addr offset

	tablrd	0,1,Xi		; auto increment for Imag Data
	tlrd	0,Xi
	tlrd	1,Xi+B1		; real data XI
	tablrd	0,0,Yi
	tlrd	0,Yi
	tlrd	1,Yi+B1		; imag data YI
;
; Real & Imag Data is fetched
; Compute Butterfly 
;
	SUB16ACC	Xl,Xi,Xt	; Xt = Xi - Xl
	ADD16		Xl,Xi		; Xi = Xi + Xl
	SUB16ACC	Yl,Yi,Yt	; Yt = Yi - Yl
	ADD16		Yl,Yi		; Yi = Yi + Yl
;
	#if SCALE_BUTTERFLY
		RRC16	Xi
		RRC16	Yi
		RRC16	Xt
		RRC16	Yt
	#endif
;
	
	MOVFP16	Cos,AARG
	MOVFP16	Yt,BARG
	Call	DblMult		; COS*Yt
	MOVPF32	DPX,ACC

	MOVFP16	Sin,AARG
	MOVFP16	Xt,BARG
	Call	DblMult		; SIN*Xt, Scale if necessary

	ADD32	ACC,DPX			
	MOVPF16	DPX+B2,Yl	; Yl = COS*Yt + SIN*Xt, Scale if necessary
;
	MOVFP16	Yt,BARG		; AARG = SIN, BARG = Yt
	Call	DblMult		; SIN*Yt
	MOVPF32	DPX,ACC

	MOVFP16	Cos,AARG
	MOVFP16	Xt,BARG
	Call	DblMult		; COS*Xt, Scale if necessary

	SUB32	ACC,DPX		; DPX = COS*Xt - SIN*Yt
	MOVPF16	DPX+B2,Xl	; Xl = COS*Xt - SIN*Yt, Scale if necessary
;
;
; Store results of butterfly
;
	DEC16	tblptrl		; table pointer already loaded with I addr

	tlwt	0,Xi
	tablwt	1,1,Xi+B1		; auto increment for Imag Data
	tlwt	0,Yi
	tablwt	1,0,Yi+B1		; Xi & Yi stored

	MOVFP16	VarL,tblptrl		; read data(L)
	ADDLBL	ExtRamAddr,tblptrl	; add data addr offset

	tlwt	0,Xl
	tablwt	1,1,Xl+B1		; auto increment for Imag Data
	tlwt	0,Yl
	tablwt	1,0,Yl+B1		; X(L) & Y(L) stored
;
; Increment for next Iloop
;
	RLC16AB	count1,temp		; temp = count1*2
	ADD16	temp,VarIloop		; I = I + temp
;

	MOVK16	(FftLen*2),temp
	SUB16	VarIloop,temp		; temp = 2*FftLen - I
	DEC16	temp
	btfss	temp+B1,MSB		
	goto	Iloop			; while I < 2*FftLen
;
; I Loop end
;
; increment for next J Loop
;

	INC16	VarJloop		; J = J + 1

	MOV16	count2,temp
	SUB16	VarJloop,temp		; temp = count2 - J
	DEC16	temp
	btfss	temp+B1,MSB		
	goto	Jloop			; while J < count2
;
; J Loop end
;
; increment for next K Loop
;
	RLC16	TF_Offset		; TF_Offset = 2 * TF_Offset
	decfsz	VarKloop
	goto	Kloop			; while K < Power
;
	return			; FFT complete
;
; K Loop End
; FFT Computation Over with data scrambled
; Descramble the data using "Unscramble" Routine
;
;******************************************************************
;	Unscramble Data Order Sequence
;	A digit reverse counter
;******************************************************************

	include	"reverse.asm"

;*****************************************************************
;	Include Double Precision Multiplication Routine
;*****************************************************************
SIGNED	equ	TRUE

	include	"17c42mpy.mac"

;
;*****************************************************************
;		Sine-Cosine Tables
;*****************************************************************
;
	include	"fft256.tbl"
;
;*****************************************************************
;	FFT Input/Output Data Stored In External RAM
; Operate Processor In Extended Microcontroller Mode
; External Data Starts at Address 0x0800, with 2 bytes of
; Real Data followed by 2 bytes of Imaginary Data.
;*****************************************************************
	ORG	EXT_RAM_START_ADDR
;
ExtRamAddr
;
	END


